use std::{result, error, fmt, io};
use std::string::FromUtf8Error;
use std::borrow::Cow;

pub type Result<T> = result::Result<T, Error>;

#[derive(Debug)]
/// `Error` just neatly wraps the other errors used in the import.
pub enum Error {
    Io(io::Error),
    FromUtf8(FromUtf8Error),
    Other(String),
}

impl From<io::Error> for Error {
    fn from(err: io::Error) -> Error {
        Error::Io(err)
    }
}

impl From<FromUtf8Error> for Error {
    fn from(err: FromUtf8Error) -> Error {
        Error::FromUtf8(err)
    }
}

impl<'a> From<&'a str> for Error {
    fn from(err: &'a str) -> Error {
        Error::Other(err.to_owned())
    }
}

impl<'a> From<Cow<'a, str>> for Error {
    fn from(err: Cow<'a, str>) -> Error {
        Error::Other(err.into_owned())
    }
}

impl fmt::Display for Error {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        match *self {
            Error::Io(ref err) => write!(f, "io error: {}", err),
            Error::FromUtf8(ref err) => write!(f, "from ut8 error: {}", err),
            Error::Other(ref err) => write!(f, "error: {}", err),
        }
    }
}

impl error::Error for Error {
    fn description(&self) -> &str {
        match *self {
            Error::Io(ref err) => err.description(),
            Error::FromUtf8(ref err) => err.description(),
            Error::Other(ref err) => err,
        }
    }

    fn cause(&self) -> Option<&error::Error> {
        match *self {
            Error::Io(ref err) => Some(err),
            Error::FromUtf8(ref err) => Some(err),
            Error::Other(_) => None,
        }
    }
}
